#include "JSFramework.h"
#include "OpenGL.h"
#include <GL/glfw.h>
#include "Loader.h"
#include "keymap.cpp"
#include "Console.h"
#include "SCQuadTree.h"
#include "Globals.h"

#include <cstdio>
#include <map>
#include <string>
#include <fstream>
#include <AL/al.h>
#include <AL/alut.h>

using namespace std;

std::vector<JSAction> JSApp::Mappings;
int JSApp::Mouse_x, JSApp::Mouse_y, JSApp::Mouse_z;
float JSApp::MoveSpeed=100;
bool JSApp::CursorFree=0;
bool JSApp::ConsoleOpen=1;
CameraGL JSApp::Camera;
map<string, int> JSApp::Config;
RDTimer JSApp::Timer;
int JSApp::AxisState[3];	
float JSApp::FrameTime=0;	
char JSApp::KeyState[400];
bool JSApp::Paused=0;
bool JSApp::Quit=0;
int JSApp::numTriangles=0;

void JSApp::MouseBtnCallback(int btn, int state) {
	KeyState[350+btn]=state;
}
void JSApp::MouseMoveCallback(int x, int y) {
	AxisState[0]=x-Mouse_x;
	AxisState[1]=y-Mouse_y;
	Mouse_x=x; Mouse_y=y;
}
void JSApp::MouseWheelCallback(int pos) {
	AxisState[2]=pos-Mouse_z;
	Mouse_z=pos;
}
void JSApp::ResizeCallback(int width, int height) {
	glfwSetWindowSize(Config["ScreenX"], Config["ScreenY"]);
}

void JSApp::KeyCallback(int key, int state) {
	if (!state) {KeyState[key]=0; return;}
	else {
		bool ConsoleHandledKey=0;
		if (ConsoleOpen) ConsoleHandledKey=Console::ReceiveKey(key);
		if (!ConsoleHandledKey) KeyState[key]=1;
	}
}

void JSApp::Init() {
	memset(KeyState, 0, sizeof(KeyState));
	if (ReadConfig()) {
		printf("Could not read configuration file, run config.exe!\n");
		return;
	}

	if (!glfwInit()) {
		printf("GLFW failed to initialize!\n");
		return;
	}
	else printf("GLFW initialized\n");

	if(!glfwOpenWindow(Config["ScreenX"],Config["ScreenY"],
		Config["ColorBits"]/4, Config["ColorBits"]/4, Config["ColorBits"]/4, Config["ColorBits"]/4,
		Config["DepthBits"],0, (Config["Fullscreen"]) ? GLFW_FULLSCREEN : GLFW_WINDOW)) 
	{
		printf("Could not set Video Mode\n");
		glfwTerminate(); 
		return;
	}
	glfwSetWindowSizeCallback(ResizeCallback);

	int ActualDepth=Config["DepthBits"], ActualColor=Config["ColorBits"];
	Console::Output("OpenGL initialized\n");
	char out[50];
	_snprintf(out, 50, "Video Mode set to %dbit Color and %dbit Depth\n", ActualColor, ActualDepth);
	Console::Output(out);

	if (InitExtensions()) Console::Output("ERROR: OpenGL extenstions failed to initialize!\n");
	else Console::Output("OpenGL extensions initialized\n");
	
	MaterialManager::Init();
	Console::Output("Material Manager initialized\n");
	
	Console::LoadBackground("console.dds");
	glViewport(0,0,Config["ScreenX"],Config["ScreenY"]);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glClearColor(0,0,0,1);

	alutInit(0, 0);
	if (alGetError()) Console::Output("ERROR: OpenAL failed to initialize!\n");
	else Console::Output("OpenAL initialized\n");

	Font::Init("verdana.glf");
	Console::Output("Font initialized\n");

	RDTimer::Init();
	Console::Output(out);
	Timer.Start();
	Console::Output("Timer initialized, Measured CPU speed: ");
	_snprintf(out, 50, "%.2fMhz\n", (0.000001f/RDTimer::cpuSpeed));
	Console::Output(out);

	Camera.setPerspective(Config["ScreenX"], Config["ScreenY"], 45, .5, 4000);
	Console::Output("Camera set to 45 degree fov and 5000 units view distance\n");

	glfwGetMousePos(&Mouse_x, &Mouse_y);
	Mouse_z=glfwGetMouseWheel();
	glfwSetKeyCallback(KeyCallback);
	glfwSetMouseButtonCallback(MouseBtnCallback);
	glfwSetMousePosCallback(MouseMoveCallback);
	glfwSetMouseWheelCallback(MouseWheelCallback);
	glfwEnable(GLFW_KEY_REPEAT);
	glfwDisable(GLFW_MOUSE_CURSOR);
	
	Console::Output("Mouse and Keyboard Callbacks registered\n");

	Console::LoadBackground("console.dds");
	Console::ExecuteScript("autoexec.csf");
	Console::ListCommands();
}

void JSApp::Shutdown() {
	Console::Output("Shutting down...\n");
	Font::Cleanup();
	alutExit();
}

int JSApp::ReadConfig() {
	Config.clear();
	ifstream file("config.cfg");
	if (!file) return -1;
	
	int value;
	string option;
	string key;

	file>>option;
	while (!file.eof()) {
		if (option == "[VIDEO]" || option == "[TERRAIN]") {
			while (!file.eof() && '['!= (file>>option, option[0])) {
				file >> value;
				Config[option]=value;
			}
		}
		else if (option == "[MAPPING]") {
			while (!file.eof() && '['!= (file>>option, option[0])) {				
				JSAction action;
				file >> key;
				action.Name=option;
				action.Key=InputToInt(key);
				Mappings.push_back(action);
			}
		}
		else file>>option;
	}
	file.close();
	return 0;
}

void JSApp::HandleActions() {
	float CamMove[3]={0};
	float CamRotate[3]={0};

	for (vector<JSAction>::iterator it = Mappings.begin(); it!=Mappings.end(); ++it) {
		if (it->Name=="Quit" && KeyState[it->Key]) {Quit=1; KeyState[it->Key]=0; return;}
		if (it->Name=="Console" && KeyState[it->Key]) {
			ConsoleOpen=!ConsoleOpen;
			KeyState[it->Key]=0;
			continue;
		}
		if (it->Name=="FreeCursor" && KeyState[it->Key]) {
			if (CursorFree=!CursorFree) glfwEnable(GLFW_MOUSE_CURSOR);
			else glfwDisable(GLFW_MOUSE_CURSOR);
			KeyState[it->Key]=0;
			continue;
		}
		if (it->Name=="Pause" && KeyState[it->Key]) {
			Paused=!Paused;
			KeyState[it->Key]=0;
			continue;
		}

		if (it->Name=="CamMoveRight" && KeyState[it->Key]) {CamMove[0]+=1; continue;}
		if (it->Name=="CamMoveLeft" && KeyState[it->Key]) {CamMove[0]+=-1; continue;}
		if (it->Name=="CamMoveForward" && KeyState[it->Key]) {CamMove[2]+=1; continue;}
		if (it->Name=="CamMoveBackward" && KeyState[it->Key]) {CamMove[2]+=-1; continue;}
		if (it->Name=="CamMoveUp" && KeyState[it->Key]) {CamMove[1]+=1; continue;}
		if (it->Name=="CamMoveDown" && KeyState[it->Key]) {CamMove[1]+=-1; continue;}
		if (it->Name=="CamMoveDown" && KeyState[it->Key]) {CamMove[1]+=-1; continue;}
		
		if (!CursorFree) {
			if (it->Name=="CamPanLR" && AxisState[it->Key]) {
				CamRotate[0]+= 0.1f * AxisState[it->Key]; continue;
			}
			if (it->Name=="CamPanUD" && AxisState[it->Key]) {
				CamRotate[1]+= 0.1f * AxisState[it->Key]; continue;
			}
			if (it->Name=="CamRoll" && AxisState[it->Key]) {
				CamRotate[2]+= 0.1f * AxisState[it->Key]; continue;
			}
		}
	}
	if (CamRotate[0]) Camera.rotate(CamRotate[0], 0,1,0, 1);
	if (CamRotate[1]) Camera.rotate(CamRotate[1], 1,0,0);
	if (CamRotate[2]) Camera.rotate(CamRotate[2], 0,0,1);
	Camera.move(CamMove[0]*FrameTime*MoveSpeed, CamMove[1]*FrameTime*MoveSpeed, CamMove[2]*FrameTime*MoveSpeed);
	memset(AxisState, 0, sizeof(AxisState));
}


void JSApp::SetMoveSpeed(float s, string unit) {
	if (unit=="kmh") s*=0.277777;
	else if (unit=="mph") s*=0.44704;
	else if (unit!="mps") Console::Output("Unknown unit, defaulting to mps (use 'kmh', 'mph' or 'mps')!\n");
	MoveSpeed=s;
	char out[100];
	_snprintf(out, 100, "Movement speed set to %.2fmps = %.2fkmh = %.2fmph\n", s, s*3.6, s*2.2369363);
	Console::Output(out);
}

void JSApp::Tick(bool newFrame) {
	if (!glfwGetWindowParam(GLFW_OPENED)) {Quit=1; return;}
	FrameTime=Timer.Reset();
	HandleActions();
	if (!newFrame) return;

	glfwSwapBuffers();
	Camera.setView();
	if (Globals::ExtCam) {
		CameraGL tmp(200,400,-150);
		tmp.rotate(-50, 0,1,0);
		tmp.rotate(50, 1,0,0);
		tmp.setView();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}
	else glClear(GL_DEPTH_BUFFER_BIT);
}

void JSApp::EnterHUDMode() {
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	float PMtr[16]={.02f,0,0,0,	 0,-.02f,0,0,  0,0,-1,0,  -1,1,0,1};
	glLoadMatrixf(PMtr);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	MaterialManager::BlendMode(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DEPTH_TEST);
}

void JSApp::LeaveHUDMode() {
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	MaterialManager::DisableBlending();
}

void JSApp::ShowStats() {
	static RDTimer FpsTimer;
	static float frames=0;
	static char strg[80]="Fps: measuring";
	static float fps=0;

	++frames;
	if (FpsTimer.Peek() > 1) {
		fps=frames/FpsTimer.Reset();
		frames=0;
	}

	if (Globals::ExtCam) Camera.DrawFrustum();
	EnterHUDMode();
	if (Globals::DrawFps) {
		_snprintf(strg, 80, "Fps: %.2f   Tris: %d  (%.2f MTri/s)", 
			fps, numTriangles, (fps*numTriangles)/1000000);
		
	}Font::Print(strg, 0.5f, .5f, 2);
	if (Globals::DrawCounters) ProfileCounter::DrawCounters();
	if (ConsoleOpen) Console::Render();
	LeaveHUDMode();
}
