#ifndef MATRIX_INCLUDED
#define MATRIX_INCLUDED

#include "VectorX.h"

class Matrix33;
class Matrix43;
class Matrix44;
static unsigned MatrixTransformationLimit=50;

class Matrix33 {
	unsigned numTrans;
public:
	Vector3 X;
	Vector3 Y;
	Vector3 Z;

	Matrix33() :numTrans(0) {};
	Matrix33(float v0, float v1, float v2, float v3, float v4, float v5, float v6, float v7, float v8)
	: numTrans(0), X(v0,v1,v2), Y(v3,v4,v5), Z(v6,v7,v8) {}

	inline void SetIdentity() {
		memset(&X, 0, 9*sizeof(float));
		X.x=Y.y=Z.z=1;
	}

	inline void ReOrtho() {
		X.normalize();
		Y.normalize();
		Z.normalize();
		numTrans=0;
	}

	inline Matrix33& RotateX(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector3 tmp=Y;
		Y.x=Y.x*cr + Z.x*sr;  Y.y=Y.y*cr + Z.y*sr;  Y.z=Y.z*cr + Z.z*sr;
		Z.x=tmp.x*-sr + Z.x*cr;  Z.y=tmp.y*-sr + Z.y*cr;  Z.z=tmp.z*-sr + Z.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix33& RotateY(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector3 tmp=X;
		X.x=X.x*cr + Z.x*sr;  X.y=X.y*cr + Z.y*sr;  X.z=X.z*cr + Z.z*sr;
		Z.x=tmp.x*-sr + Z.x*cr;  Z.y=tmp.y*-sr + Z.y*cr;  Z.z=tmp.z*-sr + Z.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix33& RotateZ(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector3 tmp=X;
		X.x=X.x*cr + Y.x*sr;  X.y=X.y*cr + Y.y*sr;  X.z=X.z*cr + Y.z*sr;
		Y.x=tmp.x*-sr + Y.x*cr;  Y.y=tmp.y*-sr + Y.y*cr;  Y.z=tmp.z*-sr + Y.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix33& TransposeIn() {
		std::swap(X.y,Y.x);
		std::swap(X.z,Z.x);
		std::swap(Y.z,Z.y);
		return *this;
	}

	inline Matrix33 Transpose() {
		return Matrix33(X.x,Y.x,Z.x, X.y,Y.y,Z.y, X.z,Y.z,Z.z);
	}

	inline Matrix33 operator*(const Matrix33& b) {
		return Matrix33(
			X.x*b.X.x + Y.x*b.X.y + Z.x*b.X.z,
			X.y*b.X.x + Y.y*b.X.y + Z.y*b.X.z,
			X.z*b.X.x + Y.z*b.X.y + Z.z*b.X.z,

			X.x*b.Y.x + Y.x*b.Y.y + Z.x*b.Y.z,
			X.y*b.Y.x + Y.y*b.Y.y + Z.y*b.Y.z,
			X.z*b.Y.x + Y.z*b.Y.y + Z.z*b.Y.z,

			X.x*b.Z.x + Y.x*b.Z.y + Z.x*b.Z.z,
			X.y*b.Z.x + Y.y*b.Z.y + Z.y*b.Z.z,
			X.z*b.Z.x + Y.z*b.Z.y + Z.z*b.Z.z);
	}

	inline Matrix33& operator*=(const Matrix33& b) {
		float tx=X.x, ty=Y.x;
		X.x=tx*b.X.x + ty*b.X.y + Z.x*b.X.z;
		Y.x=tx*b.Y.x + ty*b.Y.y + Z.x*b.Y.z;
		Z.x=tx*b.Z.x + ty*b.Z.y + Z.x*b.Z.z;

		tx=X.y; ty=Y.y;
		X.y=tx*b.X.x + ty*b.X.y + Z.y*b.X.z;
		Y.y=tx*b.Y.x + ty*b.Y.y + Z.y*b.Y.z;
		Z.y=tx*b.Z.x + ty*b.Z.y + Z.y*b.Z.z;

		tx=X.z; ty=Y.z;
		X.z=tx*b.X.x + ty*b.X.y + Z.z*b.X.z;
		Y.z=tx*b.Y.x + ty*b.Y.y + Z.z*b.Y.z;
		Z.z=tx*b.Z.x + ty*b.Z.y + Z.z*b.Z.z;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline static Matrix33 RotationX(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix33(1,0,0,
			            0,cr,sr,
						0,-sr,cr);
	}
	inline static Matrix33 RotationY(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix33(cr,0,sr,
			            0,1,0,
						-sr,0,cr);
	}
	inline static Matrix33 RotationZ(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix33(cr,sr,0,
			            -sr,cr,0,
						0,0,1);
	}
};

class Matrix43 {
public:
	float x,y,z;

	Matrix43() {};
	//inline Matrix43  operator*(const Matrix43& b) {}
	//inline Matrix43& operator*=(const Matrix43& b) {}
};

class Matrix44 {
	unsigned numTrans;
public:
	Vector4 X;
	Vector4 Y;
	Vector4 Z;
	Vector4 P;

	Matrix44() :numTrans(0) {};
	Matrix44(float v0, float v1, float v2, float v3, float v4, float v5, float v6, float v7,
		     float v8, float v9, float v10, float v11, float v12, float v13, float v14, float v15)
	: numTrans(0), X(v0,v1,v2,v3), Y(v4,v5,v6,v7), Z(v8,v9,v10,v11), P(v12,v13,v14,v15) {}

	inline void SetIdentity() {
		memset(&X.x, 0, 16*sizeof(float));
		X.x=Y.y=Z.z=P.w=1;
	}

	inline void Random() {
		float* t=&X.x;
		for (int i=0; i<16; ++i) *(t++)=(float)(rand()%10000)/5000 - 1.0;
	}

	inline void ReOrtho() {
		X.normalize();
		Y.normalize();
		Z.normalize();
		numTrans=0;
	}

	inline operator float*() {return &X.x;}

	inline Matrix44& RotateX(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector4 tmp=Y;
		Y.x=Y.x*cr + Z.x*sr;  Y.y=Y.y*cr + Z.y*sr;  Y.z=Y.z*cr + Z.z*sr;
		Z.x=tmp.x*-sr + Z.x*cr;  Z.y=tmp.y*-sr + Z.y*cr;  Z.z=tmp.z*-sr + Z.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix44& RotateY(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector4 tmp=X;
		X.x=X.x*cr + Z.x*sr;  X.y=X.y*cr + Z.y*sr;  X.z=X.z*cr + Z.z*sr;
		Z.x=tmp.x*-sr + Z.x*cr;  Z.y=tmp.y*-sr + Z.y*cr;  Z.z=tmp.z*-sr + Z.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix44& RotateZ(float rad) {
		float cr=cos(rad), sr=sin(rad);
		Vector4 tmp=X;
		X.x=X.x*cr + Y.x*sr;  X.y=X.y*cr + Y.y*sr;  X.z=X.z*cr + Y.z*sr;
		Y.x=tmp.x*-sr + Y.x*cr;  Y.y=tmp.y*-sr + Y.y*cr;  Y.z=tmp.z*-sr + Y.z*cr;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline Matrix44& TransposeIn() {
		std::swap(X.y,Y.x);
		std::swap(X.z,Z.x);
		std::swap(X.w,P.x);		
		std::swap(Y.z,Z.y);
		std::swap(Y.w,P.y);
		std::swap(Z.w,P.z);
		return *this;
	}

	inline Matrix44 Transpose() {
		return Matrix44(X.x,Y.x,Z.x,P.x, X.y,Y.y,Z.y,P.y, X.z,Y.z,Z.z,P.z, X.w,Y.w,Z.w,P.w);
	}

	inline Matrix44 operator*(const Matrix44& b) {
		Matrix44 r;
		r.X.x=X.x*b.X.x + Y.x*b.X.y + Z.x*b.X.z + P.x*b.X.w;
		r.Y.x=X.x*b.Y.x + Y.x*b.Y.y + Z.x*b.Y.z + P.x*b.Y.w;
		r.Z.x=X.x*b.Z.x + Y.x*b.Z.y + Z.x*b.Z.z + P.x*b.Z.w;
		r.P.x=X.x*b.P.x + Y.x*b.P.y + Z.x*b.P.z + P.x*b.P.w;

		r.X.y=X.y*b.X.x + Y.y*b.X.y + Z.y*b.X.z + P.y*b.X.w;
		r.Y.y=X.y*b.Y.x + Y.y*b.Y.y + Z.y*b.Y.z + P.y*b.Y.w;
		r.Z.y=X.y*b.Z.x + Y.y*b.Z.y + Z.y*b.Z.z + P.y*b.Z.w;
		r.P.y=X.y*b.P.x + Y.y*b.P.y + Z.y*b.P.z + P.y*b.P.w;

		r.X.z=X.z*b.X.x + Y.z*b.X.y + Z.z*b.X.z + P.z*b.X.w;
		r.Y.z=X.z*b.Y.x + Y.z*b.Y.y + Z.z*b.Y.z + P.z*b.Y.w;
		r.Z.z=X.z*b.Z.x + Y.z*b.Z.y + Z.z*b.Z.z + P.z*b.Z.w;
		r.P.z=X.z*b.P.x + Y.z*b.P.y + Z.z*b.P.z + P.z*b.P.w;

		r.X.w=X.w*b.X.x + Y.w*b.X.y + Z.w*b.X.z + P.w*b.X.w;
		r.Y.w=X.w*b.Y.x + Y.w*b.Y.y + Z.w*b.Y.z + P.w*b.Y.w;
		r.Z.w=X.w*b.Z.x + Y.w*b.Z.y + Z.w*b.Z.z + P.w*b.Z.w;
		r.P.w=X.w*b.P.x + Y.w*b.P.y + Z.w*b.P.z + P.w*b.P.w;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return r;
	}

	inline Matrix44& operator*=(const Matrix44& b) {
		float tx=X.x, ty=Y.x, tz=Z.x;
		X.x=tx*b.X.x + ty*b.X.y + tz*b.X.z + P.x*b.X.w;
		Y.x=tx*b.Y.x + ty*b.Y.y + tz*b.Y.z + P.x*b.Y.w;
		Z.x=tx*b.Z.x + ty*b.Z.y + tz*b.Z.z + P.x*b.Z.w;
		P.x=tx*b.P.x + ty*b.P.y + tz*b.P.z + P.x*b.P.w;

		tx=X.y; ty=Y.y; tz=Z.y;
		X.y=tx*b.X.x + ty*b.X.y + tz*b.X.z + P.y*b.X.w;
		Y.y=tx*b.Y.x + ty*b.Y.y + tz*b.Y.z + P.y*b.Y.w;
		Z.y=tx*b.Z.x + ty*b.Z.y + tz*b.Z.z + P.y*b.Z.w;
		P.y=tx*b.P.x + ty*b.P.y + tz*b.P.z + P.y*b.P.w;

		tx=X.z; ty=Y.z; tz=Z.z;
		X.z=tx*b.X.x + ty*b.X.y + tz*b.X.z + P.z*b.X.w;
		Y.z=tx*b.Y.x + ty*b.Y.y + tz*b.Y.z + P.z*b.Y.w;
		Z.z=tx*b.Z.x + ty*b.Z.y + tz*b.Z.z + P.z*b.Z.w;
		P.z=tx*b.P.x + ty*b.P.y + tz*b.P.z + P.z*b.P.w;

		tx=X.w; ty=Y.w; tz=Z.w;
		X.w=tx*b.X.x + ty*b.X.y + tz*b.X.z + P.w*b.X.w;
		Y.w=tx*b.Y.x + ty*b.Y.y + tz*b.Y.z + P.w*b.Y.w;
		Z.w=tx*b.Z.x + ty*b.Z.y + tz*b.Z.z + P.w*b.Z.w;
		P.w=tx*b.P.x + ty*b.P.y + tz*b.P.z + P.w*b.P.w;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}	

	inline Matrix44& operator*=(const Matrix33& b) {
		float tx=X.x, ty=Y.x;
		X.x=tx*b.X.x + ty*b.X.y + Z.x*b.X.z;
		Y.x=tx*b.Y.x + ty*b.Y.y + Z.x*b.Y.z;
		Z.x=tx*b.Z.x + ty*b.Z.y + Z.x*b.Z.z;

		tx=X.y; ty=Y.y;
		X.y=tx*b.X.x + ty*b.X.y + Z.y*b.X.z;
		Y.y=tx*b.Y.x + ty*b.Y.y + Z.y*b.Y.z;
		Z.y=tx*b.Z.x + ty*b.Z.y + Z.y*b.Z.z;

		tx=X.z; ty=Y.z;
		X.z=tx*b.X.x + ty*b.X.y + Z.z*b.X.z;
		Y.z=tx*b.Y.x + ty*b.Y.y + Z.z*b.Y.z;
		Z.z=tx*b.Z.x + ty*b.Z.y + Z.z*b.Z.z;
		if (++numTrans>MatrixTransformationLimit) ReOrtho();
		return *this;
	}

	inline static Matrix44 RotationX(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix44(1,0,0,0,
			            0,cr,sr,0,
						0,-sr,cr,0,
						0,0,0,1);
	}
	inline static Matrix44 RotationY(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix44(cr,0,sr,0,
			            0,1,0,0,
						-sr,0,cr,0,
						0,0,0,1);
	}
	inline static Matrix44 RotationZ(float rad) {
		float cr=cos(rad), sr=sin(rad);
		return Matrix44(cr,sr,0,0,
			            -sr,cr,0,0,
						0,0,1,0,
						0,0,0,1);
	}
};

#endif
