#ifndef TERRAIN_INCLUDED
#define TERRAIN_INCLUDED

#include "OpenGL.h"
#include "Console.h"
#include "TerrainPatch.h"
#include "SCQuadTree.h"
#include "CameraGL.h"
#include "Loader.h"
#include "Material.h"
#include "VBuffer.h"

#include <vector>

class Terrain {
private:
	struct Tile {int Size, Offset;};
	struct Linker {Tile* LowLvl;};
	struct Level {Tile Body[16]; Linker Link[4];};

	CameraGL& Camera;
	SCQuadTree QTree;
	TerrainPatch* Patches;
	TerrainPatch* RenderQ;
		
	unsigned VertexProg;
	unsigned FragmentProg;
	unsigned WaterFrgProg;
	unsigned WaterVrtProg;
	unsigned BaseTexture;
	unsigned BlendTexture;
	unsigned SphereTexture;
	unsigned DetailTexture;
	unsigned WaterNormal;
	unsigned SkyCube;

	float LodThreshold;

	Buffer** Buffers;
	Buffer* VBuffer;
	Buffer* IBuffer;
	Level* Lvl;

	int PatchSize;
	int numLevels;
	int numBuffers;
	int LOD;
	bool Available;
	bool Unaligned;

	void CreateRenderQ(unsigned node);
	void CreateTerrain();
	void CalcErrors();
	void CreateIndexBuffers();
	int BuildLinker(int level, int side, int low, unsigned short* buffer);
	inline void AddPatch2Q(TerrainPatch* Patch) {
		Patch->next=RenderQ; RenderQ=Patch;
		if (!UseDynLod) {Patch->LOD=LOD; return;}
		Vector3 d((Patch->x - (MapSize>>1)) - Camera.Position[0],
				  Patch->y - Camera.Position[1],
				  (Patch->z - (MapSize>>1)) - Camera.Position[2]);
		if (LodDistanceBased) {
			int D=100;
			float dist=d.lensq();			
			Patch->LOD=numLevels-1;
			for (int i=0; i<numLevels; ++i) {
				if (dist < D*D) {Patch->LOD=i; break;}
				D+=(int)(LodStep*(i+2));
			}
			return;
		}
		float dist=d.len();
		if (Patch->LOD && Patch->Error[Patch->LOD] > LodThreshold*dist) --Patch->LOD;
		else if (Patch->LOD < numLevels-1 && Patch->Error[Patch->LOD+1] < LodThreshold*dist) ++Patch->LOD;
	}
	int DrawTree();
	void Cleanup();

public:
	int MapSize;
	bool DrawWater;
	bool UseDynLod;
	bool LodDistanceBased;
	float LodStep;
	unsigned short* Heightmap;
		
	Terrain(CameraGL& Cam) 
	:Camera(Cam), Patches(0), RenderQ(0), LodThreshold(.45f), Buffers(0), VBuffer(0), IBuffer(0),
	Lvl(0), numLevels(0), numBuffers(0), LOD(0), Available(0), DrawWater(1), 
	UseDynLod(1), LodDistanceBased(0), LodStep(25), Heightmap(0) {
		Unaligned = (!JSApp::Config["UseVBO"] || JSApp::Config["UnalignedVBO"]) ? 1:0;
		/*if (allegro_gl_is_extension_supported("GL_ARB_fragment_program")) */Available=1;
		/*else {
			Console::Output("ERROR: GL_ARB_fragment_program not available, Terrain inactive.\n");
			return;
		}*/
		VertexProg=Loader::LoadShader("terrain.vsh");
		FragmentProg=MaterialManager::LoadShader("terrain.psh");
	}
	~Terrain() {Cleanup();}
	void SetLOD(int l);
	void SetCamera(CameraGL& cam) {Camera=cam;}
	void NewHeightmap(int msize, int psize);
	void CreateHeightmap(int msize, int psize);
	void LoadHeightmap(const char* file, int msize, int psize);
	float GetHeight(float x, float y, int radius=0);
	void Update(float dtime) {if (Available) CreateRenderQ(0);}
	void SetBasemap(const char* filename);
	void SetLoDThreshold(float t);
	int Render();
};

#endif
