#ifndef VECTORX_INCLUDED
#define VECTORX_INCLUDED

#include <cmath>

class Vector2;
class Vector3;
class Vector4;

class Vector2 {
public:
	Vector2() {};
	Vector2(float s) : x(s), y(s) {};
	Vector2(float ix, float iy) : x(ix), y(iy) {};

	float x, y;

	operator float*() {return &x;}
	
	Vector2  operator-() {return Vector2(-x, -y);}
	Vector2  operator*(float s) {return Vector2(s*x, s*y);}
	Vector2  operator/(float s) {s=1/s; return Vector2(x*s, y*s);}
	Vector2& operator=(float s) {x=y=s; return *this;}
	Vector2& operator*=(float s) {x*=s; y*=s; return *this;}	
	Vector2& operator/=(float s) {s=1/s; x*=s; y*=s; return *this;}
	
	Vector2& operator=(const float* b) {x=b[0]; y=b[1]; return *this;}
	Vector2  operator+(const float* b) {return Vector2(x+b[0], y+b[1]);}
	Vector2  operator-(const float* b) {return Vector2(x-b[0], y-b[1]);}
	float    operator*(const float* b) {return x*b[0] + y*b[1];}
	Vector2& operator+=(const float* b) {x+=b[0]; y+=b[1]; return *this;}
	Vector2& operator-=(const float* b) {x-=b[0]; y-=b[1]; return *this;}

	Vector2  operator+(Vector2& b) {return Vector2(x+b.x, y+b.y);}	
	Vector2  operator-(Vector2& b) {return Vector2(x-b.x, y-b.y);}
	float    operator*(Vector2& b) {return x*b.x + y*b.y;}
	Vector2& operator+=(Vector2& b) {x+=b.x; y+=b.y; return *this;}
	Vector2& operator-=(Vector2& b) {x-=b.x; y-=b.y; return *this;}
	
	float lensq() {return x*x + y*y;}
	float len() {return sqrt(lensq());}
	Vector2& normalize() {return *this/=len();}

	void random(float xlim, float ylim) {
		x= xlim*(1-(float)(rand()%2000)/1000);
		y= ylim*(1-(float)(rand()%2000)/1000);
	}
};

class Vector3 {
public:
	Vector3() {};
	Vector3(float s) : x(s), y(s), z(s) {};
	Vector3(float ix, float iy, float iz) : x(ix), y(iy), z(iz) {};

	float x,y,z;

	operator float*() {return &x;}
	operator Vector2*() {return (Vector2*)&x;}
	
	Vector3  operator-() {return Vector3(-x, -y, -z);}
	Vector3  operator*(float s) {return Vector3(s*x, s*y, s*z);}
	Vector3  operator/(float s) {s=1/s; return Vector3(x*s, y*s, z*s);}
	Vector3& operator=(float s) {x=y=z=s; return *this;}
	Vector3& operator*=(float s) {x*=s; y*=s; z*=s; return *this;}	
	Vector3& operator/=(float s) {s=1/s; x*=s; y*=s; z*=s; return *this;}
	
	Vector3& operator=(const float* b) {x=b[0]; y=b[1]; z=b[2]; return *this;}
	Vector3  operator+(const float* b) {return Vector3(x+b[0], y+b[1], z+b[2]);}
	Vector3  operator-(const float* b) {return Vector3(x-b[0], y-b[1], z-b[2]);}
	float    operator*(const float* b) {return x*b[0] + y*b[1] + z*b[2];}
	Vector3  operator%(const float* b) {return Vector3(y*b[2]-z*b[1], z*b[0]-x*b[2], x*b[1]-y*b[0]);}	
	Vector3& operator+=(const float* b) {x+=b[0]; y+=b[1]; z+=b[2]; return *this;}
	Vector3& operator-=(const float* b) {x-=b[0]; y-=b[1]; z-=b[2]; return *this;}
	Vector3& operator%=(const float* b) {float tmpx=x, tmpy=y; x=y*b[2]-z*b[1]; y=z*b[0]-tmpx*b[2]; z=tmpx*b[1]-tmpy*b[0]; return *this;}

	Vector3  operator+(Vector3& b) {return Vector3(x+b.x, y+b.y, z+b.z);}	
	Vector3  operator-(Vector3& b) {return Vector3(x-b.x, y-b.y, z-b.z);}
	float    operator*(Vector3& b) {return x*b.x + y*b.y + z*b.z;}
	Vector3  operator%(Vector3& b) {return Vector3(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);}
	Vector3& operator+=(Vector3& b) {x+=b.x; y+=b.y; z+=b.z; return *this;}
	Vector3& operator-=(Vector3& b) {x-=b.x; y-=b.y; z-=b.z; return *this;}
	Vector3& operator%=(Vector3& b) {float tmpx=x, tmpy=y; x=y*b.z-z*b.y; y=z*b.x-tmpx*b.z; z=tmpx*b.y-tmpy*b.x; return *this;}
	
	float lensq() {return x*x + y*y + z*z;}
	float len() {return sqrt(lensq());}
	Vector3& normalize() {return *this/=len();}

	void random(float xlim, float ylim, float zlim) {
		x= xlim*(1-(float)(rand()%2000)/1000);
		y= ylim*(1-(float)(rand()%2000)/1000);
		z= zlim*(1-(float)(rand()%2000)/1000);
	}
};

class Vector4 {
public:
	Vector4() {};
	Vector4(float s) : x(s), y(s), z(s), w(s) {};
	Vector4(float ix, float iy, float iz, float iw) : x(ix), y(iy), z(iz), w(iw) {};

	float x,y,z,w;

	operator float*() {return &x;}
	operator Vector2*() {return (Vector2*)&x;}
	operator Vector3*() {return (Vector3*)&x;}

	Vector4  operator-() {return Vector4(-x, -y, -z, -w);}
	Vector4  operator*(float s) {return Vector4(s*x, s*y, s*z, s*w);}
	Vector4  operator/(float s) {s=1/s; return Vector4(x*s, y*s, z*s, w*s);}
	Vector4& operator=(float s) {x=y=z=w=s; return *this;}
	Vector4& operator*=(float s) {x*=s; y*=s; z*=s; w*=s; return *this;}	
	Vector4& operator/=(float s) {s=1/s; x*=s; y*=s; z*=s; w*=s; return *this;}
	
	Vector4& operator=(const float* b) {x=b[0]; y=b[1]; z=b[2]; w=b[3]; return *this;}
	Vector4  operator+(const float* b) {return Vector4(x+b[0], y+b[1], z+b[2], w+b[3]);}
	Vector4  operator-(const float* b) {return Vector4(x-b[0], y-b[1], z-b[2], w-b[3]);}
	float     operator*(const float* b) {return x*b[0] + y*b[1] + z*b[2] + w*b[3];}
	Vector4& operator+=(const float* b) {x+=b[0]; y+=b[1]; z+=b[2]; w+=b[3]; return *this;}
	Vector4& operator-=(const float* b) {x-=b[0]; y-=b[1]; z-=b[2]; w-=b[3]; return *this;}
	
	Vector4  operator+(Vector4& b) {return Vector4(x+b.x, y+b.y, z+b.z, w+b.w);}	
	Vector4  operator-(Vector4& b) {return Vector4(x-b.x, y-b.y, z-b.z, w-b.w);}
	float     operator*(Vector4& b) {return x*b.x + y*b.y + z*b.z + w*b.w;}
	Vector4& operator+=(Vector4& b) {x+=b.x; y+=b.y; z+=b.z; w+=b.w; return *this;}
	Vector4& operator-=(Vector4& b) {x-=b.x; y-=b.y; z-=b.z; w-=b.w; return *this;}
	
	float lensq() {return x*x + y*y + z*z + w*w;}
	float len() {return sqrt(lensq());}
	Vector4& normalize() {return *this/=len();}

	void random(float xlim, float ylim, float zlim, float wlim) {
		x= xlim*(1-(float)(rand()%2000)/1000);
		y= ylim*(1-(float)(rand()%2000)/1000);
		z= zlim*(1-(float)(rand()%2000)/1000);
		w= wlim*(1-(float)(rand()%2000)/1000);
	}
};

#endif
