#include "cameraGL.h"

#include <cmath>
#include <memory.h>
#include "OpenGL.h"
#include "Material.h"

CameraGL::CameraGL(float x, float y, float z) {
	memset(Transform, 0, sizeof(Transform));
	Transform[0] = Transform[5] = Transform[10] = Transform[15] = 1.0f;
	Transform[12] = x; Transform[13] = y; Transform[14] = z;

	Right=&Transform[0];
	Up=&Transform[4];
	Forward=&Transform[8];
	Position=&Transform[12];

	FrustumColor[0]=.2f;
	FrustumColor[1]=.4f;
	FrustumColor[2]=1;
}

void CameraGL::SetFrustumColor(float r, float g, float b) {
	FrustumColor[0]=r; FrustumColor[1]=g; FrustumColor[2]=b;
};

void CameraGL::setViewDistance(float n, float f) {
	setPerspective(ResX, ResY, FoV, n, f);
}

void CameraGL::setFoV(float fov) {
	setPerspective(ResX, ResY, fov, NearP, FarP);
}

void CameraGL::setPerspective(unsigned rX, unsigned rY, float fov, float nearp, float farp) {
	ResX=rX; ResY=rY;
	FoV=fov;
	NearP=nearp; FarP=farp;
	yFac = tanf(FoV * 3.141592f/360);
	xFac = yFac*(float)rX/rY;

	float dx=(float)sqrt(1.0 + xFac*xFac);
	float dy=(float)sqrt(1.0 + yFac*yFac);
	BLN[0]=1.0f/dx;	BLN[1]=xFac/dx;
	BRN[0]=-1.0f/dx; BRN[1]=xFac/dx;
	BBN[0]=1.0f/dy; BBN[1]=yFac/dy;
	BTN[0]=-1.0f/dy; BTN[1]=yFac/dy;
	
	glMatrixMode(GL_PROJECTION);
	float PMtr[16]={1/xFac, 0, 0, 0,	
					0, 1/yFac, 0, 0,
					0, 0, (farp+nearp)/(farp-nearp), 1,
					0, 0, (2*farp*nearp)/(nearp-farp), 0};
	glLoadMatrixf(PMtr);
	glMatrixMode(GL_MODELVIEW);
}

void CameraGL::setView() {
	glMatrixMode(GL_MODELVIEW);
	float viewmatrix[16]={
		Transform[0], Transform[4], Transform[8], 0,
		Transform[1], Transform[5], Transform[9], 0,
		Transform[2], Transform[6], Transform[10], 0,
		-(Transform[0]*Transform[12] + Transform[1]*Transform[13] + Transform[2]*Transform[14]),
		-(Transform[4]*Transform[12] + Transform[5]*Transform[13] +	Transform[6]*Transform[14]),						  
		-(Transform[8]*Transform[12] + Transform[9]*Transform[13] + Transform[10]*Transform[14]), 1};
	glLoadMatrixf(viewmatrix);

	LN[0]=BLN[0]*Right[0] + BLN[1]*Forward[0];
	LN[1]=BLN[0]*Right[1] + BLN[1]*Forward[1];
	LN[2]=BLN[0]*Right[2] + BLN[1]*Forward[2];

	RN[0]=BRN[0]*Right[0] + BRN[1]*Forward[0];
	RN[1]=BRN[0]*Right[1] + BRN[1]*Forward[1];
	RN[2]=BRN[0]*Right[2] + BRN[1]*Forward[2];

	TN[0]=BTN[0]*Up[0] + BTN[1]*Forward[0];
	TN[1]=BTN[0]*Up[1] + BTN[1]*Forward[1];
	TN[2]=BTN[0]*Up[2] + BTN[1]*Forward[2];

	BN[0]=BBN[0]*Up[0] + BBN[1]*Forward[0];
	BN[1]=BBN[0]*Up[1] + BBN[1]*Forward[1];
	BN[2]=BBN[0]*Up[2] + BBN[1]*Forward[2];
}

void CameraGL::move(float x, float y, float z, bool global, float distance) {
	if (global) {
		Transform[12] += x * distance;
		Transform[13] += y * distance;
		Transform[14] += z * distance;
	}
	else {
		Transform[12] += distance * (x*Transform[0] + y*Transform[4] + z*Transform[8]);
		Transform[13] += distance * (x*Transform[1] + y*Transform[5] + z*Transform[9]);
		Transform[14] += distance * (x*Transform[2] + y*Transform[6] + z*Transform[10]);
	}
}

void CameraGL::rotate(float deg, float x, float y, float z, bool global) {
	glPushMatrix();
	glLoadMatrixf(Transform);

	if (global) glRotatef(deg,
		x*Transform[0] + y*Transform[1] + z*Transform[2],
		x*Transform[4] + y*Transform[5] + z*Transform[6],
		x*Transform[8] + y*Transform[9] + z*Transform[10]);
	else glRotatef(deg, x,y,z);
	glGetFloatv(GL_MODELVIEW_MATRIX, Transform);
	glPopMatrix();
}

int CameraGL::SphereVisible(float pos[3], float radius) {
	float d[3]={pos[0]-Position[0], pos[1]-Position[1], pos[2]-Position[2]};

	float camZ=Forward[0]*d[0] + Forward[1]*d[1] + Forward[2]*d[2];
	if (camZ+radius < NearP || camZ-radius > FarP) return 0;
	
	float camXl=LN[0]*d[0] + LN[1]*d[1] + LN[2]*d[2];
	if (camXl < -radius) return 0;
	float camXr=RN[0]*d[0] + RN[1]*d[1] + RN[2]*d[2];
	if (camXr < -radius) return 0;

	float camYt=TN[0]*d[0] + TN[1]*d[1] + TN[2]*d[2];
	if (camYt < -radius) return 0;
	float camYb=BN[0]*d[0] + BN[1]*d[1] + BN[2]*d[2];
	if (camYb < -radius) return 0;

	if (camZ-radius>NearP && camZ+radius<FarP && camXl > radius &&
		camXr > radius && camYt > radius && camYb > radius) return 1;
	
	return -1;
}

bool CameraGL::BoxVisible(float bbox[24]) {
	float d[3];
	int l=0, r=0, f=0, n=0, t=0, b=0;
	bool inX, inY, inZ;
	float camX, camY, camZ;

	for (int i=0; i<8; ++i) {
		inX=inY=inZ=false;		
		d[0]=bbox[(i*3)]-Position[0];
		d[1]=bbox[(i*3)+1]-Position[1];
		d[2]=bbox[(i*3)+2]-Position[2];

		camX=Right[0]*d[0] + Right[1]*d[1] + Right[2]*d[2];
		camY=Up[0]*d[0] + Up[1]*d[1] + Up[2]*d[2];
		camZ=Forward[0]*d[0] + Forward[1]*d[1] + Forward[2]*d[2];

		if (camX < -xFac*camZ) ++l;
		else if (camX > xFac*camZ) ++r;
		else inX=true;
		
		if (camY < -yFac*camZ) b++;
		else if (camY > yFac*camZ) t++;
		else inY=true;

		if (camZ < NearP) ++n;		
		else if (camZ > FarP) ++f;
		else inZ=true;

		if (inX && inY && inZ) return true;
	}

	if (l==8 || r==8 || f==8 || n==8 || b==8 || t==8) return false;
	return true;
}

bool CameraGL::AABBVisible(float bbox[6]) {
	int l=0, r=0, f=0, n=0, t=0, b=0;
	bool inX, inY, inZ;
	float camX, camY, camZ;

	float tmpcorners[2][3];
	tmpcorners[0][0]=bbox[0]-Position[0];
	tmpcorners[0][1]=bbox[1]-Position[1];
	tmpcorners[0][2]=bbox[2]-Position[2];
	tmpcorners[1][0]=bbox[3]-Position[0];
	tmpcorners[1][1]=bbox[4]-Position[1];
	tmpcorners[1][2]=bbox[5]-Position[2];

	for (int i=0; i<8; ++i) {
		inX=inY=inZ=false;
		float d[3]={tmpcorners[i&1][0], tmpcorners[(i>>2)&1][1], tmpcorners[(i>>1)&1][2]};

		camX=Right[0]*d[0] + Right[1]*d[1] + Right[2]*d[2];
		camY=Up[0]*d[0] + Up[1]*d[1] + Up[2]*d[2];
		camZ=Forward[0]*d[0] + Forward[1]*d[1] + Forward[2]*d[2];

		if (camX < -xFac*camZ) ++l;
		else if (camX > xFac*camZ) ++r;
		else inX=true;
		
		if (camY < -yFac*camZ) b++;
		else if (camY > yFac*camZ) t++;
		else inY=true;

		if (camZ < NearP) ++n;		
		else if (camZ > FarP) ++f;
		else inZ=true;

		if (inX && inY && inZ) return true;
	}

	if (l==8 || r==8 || f==8 || n==8 || b==8 || t==8) return false;
	return true;
}

void CameraGL::DrawFrustum(float dist) {
	if (!dist) dist=FarP;
	float front[3]={Position[0]+Forward[0]*dist, Position[1]+Forward[1]*dist, Position[2]+Forward[2]*dist};
	float up[3]={Up[0]*yFac*dist, Up[1]*yFac*dist, Up[2]*yFac*dist};
	float right[3]={Right[0]*xFac*dist, Right[1]*xFac*dist, Right[2]*xFac*dist};

	MaterialManager::DisableBlending();
	MaterialManager::SetColor(FrustumColor[0]*FrustumColor[0], 
							  FrustumColor[1]*FrustumColor[1],
							  FrustumColor[2]*FrustumColor[2], 1);

	glBegin(GL_LINE_LOOP);
		glVertex3f(Position[0], Position[1], Position[2]);
		glVertex3f(front[0]+up[0]+right[0], front[1]+up[1]+right[1], front[2]+up[2]+right[2]);
		glVertex3f(front[0]+up[0]-right[0], front[1]+up[1]-right[1], front[2]+up[2]-right[2]);
		glVertex3f(Position[0], Position[1], Position[2]);
		glVertex3f(front[0]-up[0]+right[0], front[1]-up[1]+right[1], front[2]-up[2]+right[2]);
		glVertex3f(front[0]-up[0]-right[0], front[1]-up[1]-right[1], front[2]-up[2]-right[2]);
	glEnd();

	glBegin(GL_LINES);
		glVertex3f(front[0]+up[0]+right[0], front[1]+up[1]+right[1], front[2]+up[2]+right[2]);
		glVertex3f(front[0]-up[0]+right[0], front[1]-up[1]+right[1], front[2]-up[2]+right[2]);
		glVertex3f(front[0]+up[0]-right[0], front[1]+up[1]-right[1], front[2]+up[2]-right[2]);
		glVertex3f(front[0]-up[0]-right[0], front[1]-up[1]-right[1], front[2]-up[2]-right[2]);
	glEnd();

	MaterialManager::DisableTextures();
	MaterialManager::SetColor(FrustumColor[0]*2, FrustumColor[1]*2, FrustumColor[2]*2, .5f);
	MaterialManager::BlendMode(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0, 1.0);

	glBegin(GL_TRIANGLE_FAN);
	glVertex3fv(Position);
	glVertex3f(front[0]+up[0]+right[0], front[1]+up[1]+right[1], front[2]+up[2]+right[2]);
	glVertex3f(front[0]+up[0]-right[0], front[1]+up[1]-right[1], front[2]+up[2]-right[2]);
	glVertex3f(front[0]-up[0]-right[0], front[1]-up[1]-right[1], front[2]-up[2]-right[2]);
	glVertex3f(front[0]-up[0]+right[0], front[1]-up[1]+right[1], front[2]-up[2]+right[2]);
	glVertex3f(front[0]+up[0]+right[0], front[1]+up[1]+right[1], front[2]+up[2]+right[2]);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(front[0]-up[0]+right[0], front[1]-up[1]+right[1], front[2]-up[2]+right[2]);
	glVertex3f(front[0]-up[0]-right[0], front[1]-up[1]-right[1], front[2]-up[2]-right[2]);
	glVertex3f(front[0]+up[0]-right[0], front[1]+up[1]-right[1], front[2]+up[2]-right[2]);
	glVertex3f(front[0]+up[0]+right[0], front[1]+up[1]+right[1], front[2]+up[2]+right[2]);
	glEnd();

	glDisable(GL_POLYGON_OFFSET_FILL);

	MaterialManager::SetColor(1,1,1,1);
}

void CameraGL::Screen2World(float x, float y, float vec[3]) {
	x= xFac * (x*ResX)/2;
	y=-yFac * (y*ResY)/2;
	vec[0]=Forward[0] + x*Right[0] + y*Up[0];
	vec[1]=Forward[1] + x*Right[1] + y*Up[1];
	vec[2]=Forward[2] + x*Right[2] + y*Up[2];
}

void CameraGL::reset() {
	memset(Transform, 0, sizeof(Transform));
	Transform[0] = Transform[5] = Transform[10] = Transform[15] = 1.0f;
}
